/*
Module : AADATE.CPP
Purpose: Implementation for the algorithms which convert between the Gregorian and Julian calendars and the Julian Day
Created: PJN / 29-12-2003
History: PJN / 10-11-2004 1. Fix for CAADate::Get so that it works correctly for propalactive calendar dates
         PJN / 15-05-2005 1. Fix for CAADate::Set(double JD, bool bGregorianCalendar) not setting the m_bGregorianCalendar
                          member variable correctly. 

Copyright (c) 2003 - 2006 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////////////// Includes /////////////////////////////////////////

#include "stdafx.h"
#include "AADate.h"
#include <cmath>


//////////////////////////// Implementation ///////////////////////////////////

CAADate::CAADate()
{
  m_dblJulian = 0;
  m_bGregorianCalendar = true;
}

CAADate::CAADate(long Year, long Month, double Day, bool bGregorianCalendar)
{
  Set(Year, Month, Day, 0, 0, 0, bGregorianCalendar);
}

CAADate::CAADate(long Year, long Month, double Day, double Hour, double Minute, double Second, bool bGregorianCalendar)
{
  Set(Year, Month, Day, Hour, Minute, Second, bGregorianCalendar);
}

CAADate::CAADate(double JD, bool bGregorianCalendar)
{
  Set(JD, bGregorianCalendar);
}

double CAADate::Set(long Year, long Month, double Day, bool bGregorianCalendar)
{
  long Y = Year;
  long M = Month;
  if (M < 3)
  {
    Y = Y - 1;
    M = M + 12;
  }

  long A = 0;
  long B = 0;
  if (bGregorianCalendar)
  {
    A = INT(Y / 100.0);
    B = 2 - A + INT(A / 4.0);
  }

  return INT(365.25 * (Y + 4716)) + INT(30.6001 * (M + 1)) + Day + B - 1524.5;
}

bool CAADate::IsLeap(long Year, bool bGregorianCalendar)
{     
  if (bGregorianCalendar)    
  {                                       
    if ((Year % 100) == 0)                
      return ((Year % 400) == 0) ? true : false;
    else
      return ((Year % 4) == 0) ? true : false;
  }
  else                                        
    return ((Year % 4) == 0) ? true : false;
}

void CAADate::Set(long Year, long Month, double Day, double Hour, double Minute, double Second, bool bGregorianCalendar)
{
  double dblDay = Day + (Hour/24) + (Minute/1440) + (Second / 86400);

  m_bGregorianCalendar = bGregorianCalendar;
  m_dblJulian = Set(Year, Month, dblDay, m_bGregorianCalendar);
} 

void CAADate::Get(long& Year, long& Month, long& Day, long& Hour, long& Minute, double& Second) const
{
  double JD = m_dblJulian + 0.5;
  double tempZ; 
  double F = modf(JD, &tempZ);
  long Z = static_cast<long>(tempZ);
  long A = Z;
  if (m_bGregorianCalendar) //There is a difference here between the Meeus implementation and this one
  //if (Z >= 2299161)       //The Meeus implementation automatically assumes the Gregorian Calendar 
                            //came into effect on 15 October 1582 (JD: 2299161), while the CAADate
                            //implementation has a "m_bGregorianCalendar" value to decide if the date
                            //was specified in the Gregorian or Julian Calendars. This difference
                            //means in effect that CAADate fully supports propalactive versions of both
                            //calendar systems. This problem was discovered when testing the Moslem
                            //calendar code which involves a round trip between Moslem -> Julian ->
                            //Gregorian calendars. 
  {
    long alpha = INT((Z - 1867216.25) / 36524.25);
    A = Z + 1 + alpha - INT(alpha/4.0);
  }

  long B = A + 1524;
  long C = INT((B - 122.1) / 365.25);
  long D = INT(365.25 * C);
  long E = INT((B - D) / 30.6001);

  double dblDay = B - D - INT(30.6001 * E) + F;
  Day = static_cast<long>(dblDay);

  if (E < 14)
    Month = E - 1;
  else
    Month = E - 13;

  if (Month > 2)
    Year = C - 4716;
  else
    Year = C - 4715;

  F = modf(dblDay, &tempZ);
  Hour = INT(F*24);
  Minute = INT((F - (Hour)/24.0)*1440.0);
  Second = (F - (Hour / 24.0) - (Minute / 1440.0)) * 86400.0;
}

void CAADate::Set(double JD, bool bGregorianCalendar)
{
  m_dblJulian = JD;
  m_bGregorianCalendar = bGregorianCalendar;
}

long CAADate::Day() const
{
  long Year = 0;
  long Month = 0;
  long Day = 0;
  long Hour = 0;
  long Minute = 0;
  double Second = 0;
  Get(Year, Month, Day, Hour, Minute, Second);
  return Day;
}

long CAADate::Month() const
{
  long Year = 0;
  long Month = 0;
  long Day = 0;
  long Hour = 0;
  long Minute = 0;
  double Second = 0;
  Get(Year, Month, Day, Hour, Minute, Second);
  return Month;
}

long CAADate::Year() const
{
  long Year = 0;
  long Month = 0;
  long Day = 0;
  long Hour = 0;
  long Minute = 0;
  double Second = 0;
  Get(Year, Month, Day, Hour, Minute, Second);
  return Year;
}

long CAADate::Hour() const
{
  long Year = 0;
  long Month = 0;
  long Day = 0;
  long Hour = 0;
  long Minute = 0;
  double Second = 0;
  Get(Year, Month, Day, Hour, Minute, Second);
  return Hour;
}

long CAADate::Minute() const
{
  long Year = 0;
  long Month = 0;
  long Day = 0;
  long Hour = 0;
  long Minute = 0;
  double Second = 0;
  Get(Year, Month, Day, Hour, Minute, Second);
  return Minute;
}

double CAADate::Second() const
{
  long Year = 0;
  long Month = 0;
  long Day = 0;
  long Hour = 0;
  long Minute = 0;
  double Second = 0;
  Get(Year, Month, Day, Hour, Minute, Second);
  return Second;
}

CAADate::DAY_OF_WEEK CAADate::DayOfWeek() const
{
  return static_cast<DAY_OF_WEEK>((static_cast<long>(m_dblJulian + 1.5) % 7));
}

long CAADate::DaysInMonth() const
{
  long Year = 0;
  long Month = 0;
  long Day = 0;
  long Hour = 0;
  long Minute = 0;
  double Second = 0;
  Get(Year, Month, Day, Hour, Minute, Second);

  int MonthLength[14] =
  { 
    31, 28, 31, 30, 31, 30,
    31, 31, 30, 31, 30, 31, 0
  };
  if (IsLeap(Year, m_bGregorianCalendar))
    MonthLength[1]++;
  
  return MonthLength[Month-1];
}

long CAADate::DaysInYear() const
{
  long Year = 0;
  long Month = 0;
  long Day = 0;
  long Hour = 0;
  long Minute = 0;
  double Second = 0;
  Get(Year, Month, Day, Hour, Minute, Second);

  if (IsLeap(Year, m_bGregorianCalendar))
    return 366;
  else
    return 365;
}

double CAADate::DayOfYear() const
{
  long Year = 0;
  long Month = 0;
  long Day = 0;
  long Hour = 0;
  long Minute = 0;
  double Second = 0;
  Get(Year, Month, Day, Hour, Minute, Second);

  return m_dblJulian - Set(Year, 1, 1, m_bGregorianCalendar) + 1;
}

double CAADate::FractionalYear() const
{
  long Year = 0;
  long Month = 0;
  long Day = 0;
  long Hour = 0;
  long Minute = 0;
  double Second = 0;
  Get(Year, Month, Day, Hour, Minute, Second);

  long DaysInYear;
  if (IsLeap(Year, m_bGregorianCalendar))
    DaysInYear = 366;
  else
    DaysInYear = 365;

  return Year + ((m_dblJulian - Set(Year, 1, 1, m_bGregorianCalendar)) / DaysInYear);
}

bool CAADate::Leap() const
{
  return IsLeap(Year(), m_bGregorianCalendar);
}

void CAADate::DayOfYearToDayAndMonth(long DayOfYear, bool bLeap, long& DayOfMonth, long& Month)
{
  long K = bLeap ? 1 : 2;

  Month = INT(9*(K + DayOfYear)/275.0 + 0.98);
  if (DayOfYear < 32)
    Month = 1;

  DayOfMonth = DayOfYear - INT((275*Month)/9.0) + (K*INT((Month + 9)/12.0)) + 30;
}

long CAADate::INT(double value)
{
  if (value >= 0)
    return static_cast<long>(value);
  else
    return static_cast<long>(value - 1);
}

CAACalendarDate CAADate::JulianToGregorian(long Year, long Month, long Day)
{
  CAADate Julian(Year, Month, Day, false);
  long J = static_cast<long>(Julian.DayOfYear());

  long JD = CAADate::INT(365.25 * (Year - 1)) + 1721423 + J;
  long alpha = CAADate::INT((JD - 1867216.25) / 36524.25);
  long beta = JD + 1 + alpha - CAADate::INT(alpha / 4.0);
  if (JD < 2299161)
    beta = JD;

  long b = beta + 1524;
  long c = CAADate::INT((b - 122.1) / 365.25);
  long d = CAADate::INT(365.25 * c);
  long e = CAADate::INT((b - d) / 30.6001);

  CAACalendarDate GregorianDate;
  GregorianDate.Day = b - d - INT(30.6001 * e);
  if (e < 14)
    GregorianDate.Month = e - 1;
  else
    GregorianDate.Month = e - 13;

  if (GregorianDate.Month > 2)
    GregorianDate.Year = c - 4716;
  else
    GregorianDate.Year = c - 4715;

  return GregorianDate;
}

CAACalendarDate CAADate::GregorianToJulian(long Year, long Month, long Day)
{
  if (Month < 3)
  {
    Year--;
    Month += 12;
  }

  long alpha = CAADate::INT(Year / 100.0);
  long beta = 2 - alpha + CAADate::INT(alpha / 4.0);
  long b = CAADate::INT(365.25 * Year) + CAADate::INT(30.6001 * (Month + 1)) + Day + 1722519 + beta;
  long c = CAADate::INT((b - 122.1) / 365.25);
  long d = CAADate::INT(365.25 * c);
  long e = CAADate::INT((b - d) / 30.6001);

  CAACalendarDate JulianDate;
  JulianDate.Day = b - d - INT(30.6001 * e);
  if (e < 14)
    JulianDate.Month = e - 1;
  else
    JulianDate.Month = e - 13;

  if (JulianDate.Month > 2)
    JulianDate.Year = c - 4716;
  else
    JulianDate.Year = c - 4715;

  return JulianDate;
}

